from dataclasses import dataclass
import json
from pathlib import Path
import sys
from typing import List, Optional

from requests import Session

from .media import AudioFormat
from .models import AlbumsEndpointResponseJSON
from .requesting import request_albums, request_album_items, request_album_review
from .track import Track
from .utils import download_cover_image


@dataclass
class Album:
    album_id: int

    def __post_init__(self):
        self.album_dir: Optional[Path] = None
        self.album_cover_saved: bool = False

    def get_items(self, session: Session):
        album_items: AlbumsItemsResponseJSON = request_album_items(
            session=session, identifier=self.album_id
        )
        _items = album_items.items if album_items is not None else ()
        self.tracks = tuple(_item.item for _item in _items)

    def get_metadata(self, session: Session):
        self.metadata: AlbumsEndpointResponseJSON = request_albums(
            session=session, identifier=self.album_id
        )

    def get_review(self, session: Session):
        if self.album_dir is None:
            self.set_dir(out_dir=out_dir)
        self.album_review: Optional[AlbumsReviewResponseJSON] = request_album_review(
            session=session, identifier=self.album_id
        )
        if self.album_review is not None:
            (self.album_dir / "AlbumReview.json").write_text(
                self.album_review.to_json()
            )

    def set_dir(self, out_dir: Path):
        artist_substring: str = self.metadata.artist.name.replace("..", "")
        album_substring: str = (
            f"{self.metadata.name.replace('..', '')} "
            f"[{self.metadata.id}] [{self.metadata.release_date.year}]"
        )
        self.album_dir = out_dir / artist_substring / album_substring
        self.album_dir.mkdir(parents=True, exist_ok=True)

        if self.metadata.number_of_volumes > 1:
            for v in range(1, self.metadata.number_of_volumes + 1):
                volume_substring: str = f"Volume {v}"
                (out_dir / artist_substring / album_substring / volume_substring).mkdir(
                    parents=True, exist_ok=True
                )

    def save_cover_image(self, session: Session, out_dir: Path):
        if self.album_dir is None:
            self.set_dir(out_dir=out_dir)
        self.cover_path: Path = self.album_dir / "cover.jpg"
        if not self.cover_path.exists():
            download_cover_image(
                session=session,
                cover_uuid=self.metadata.cover,
                output_dir=self.album_dir,
            )
        else:
            self.album_cover_saved = True

    def get_tracks(
        self, session: Session, audio_format: AudioFormat, out_dir: Path
    ) -> List[Optional[str]]:
        track_files: List[str] = [None] * self.metadata.number_of_tracks
        for i, t in enumerate(self.tracks):  # type(t) is TracksEndpointResponseJSON
            track: Track = Track(track_id=t.id)

            track_files_value: Optional[str] = track.get(
                session=session,
                audio_format=audio_format,
                out_dir=out_dir,
                metadata=t,
                album=self.metadata,
            )
            track_files[i] = {track.metadata.track_number: track_files_value}
        else:
            self.track_files = track_files

    def dumps(self):
        return json.dumps(self.track_files)

    def dump(self, fp=sys.stdout):
        json.dump(self.track_files, fp)

    def get(
        self,
        session: Session,
        audio_format: AudioFormat,
        out_dir: Path,
        metadata: Optional[AlbumsEndpointResponseJSON] = None,
    ):
        if metadata is None:
            self.get_metadata(session)
        else:
            self.metadata = metadata

        self.get_items(session)
        self.save_cover_image(session, out_dir)
        self.get_review(session)
        self.get_tracks(session, audio_format, out_dir)



from functools import partial
import logging
from typing import Callable, Dict, Iterable, Iterator, Optional, Tuple, Union

from .models import (
    AlbumsEndpointResponseJSON,
    AlbumsItemsResponseJSON,
    AlbumsReviewResponseJSON,
    ArtistsAlbumsResponseJSON,
    ArtistsBioResponseJSON,
    ArtistsEndpointResponseJSON,
    ArtistsVideosResponseJSON,
    PlaylistsEndpointResponseJSON,
    SessionsEndpointResponseJSON,
    SubscriptionEndpointResponseJSON,
    TracksCreditsResponseJSON,
    TracksEndpointResponseJSON,
    TracksEndpointStreamResponseJSON,
    TracksLyricsResponseJSON,
    VideosContributorsResponseJSON,
    VideosEndpointResponseJSON,
    VideosEndpointStreamResponseJSON,
)
from .utils import TIDAL_API_URL

import backoff
from requests import HTTPError, PreparedRequest, Request, Response, Session

logger: logging.Logger = logging.getLogger(__name__)

ResponseJSON = Union[
    AlbumsEndpointResponseJSON,
    AlbumsItemsResponseJSON,
    AlbumsReviewResponseJSON,
    ArtistsAlbumsResponseJSON,
    ArtistsBioResponseJSON,
    ArtistsEndpointResponseJSON,
    ArtistsVideosResponseJSON,
    PlaylistsEndpointResponseJSON,
    SessionsEndpointResponseJSON,
    SubscriptionEndpointResponseJSON,
    TracksCreditsResponseJSON,
    TracksEndpointResponseJSON,
    TracksEndpointStreamResponseJSON,
    TracksLyricsResponseJSON,
    VideosContributorsResponseJSON,
    VideosEndpointResponseJSON,
    VideosEndpointStreamResponseJSON,
]


def requester_maker(
    session: Session,
    endpoint: str,
    identifier: int,
    url_end: str = "",
    headers: Optional[dict] = None,
    parameters: Optional[dict] = None,
    subclass: Optional[ResponseJSON] = None,
    credits_flag: bool = False,
) -> Callable:
    """This function is a function factory: it crafts nearly identical
    versions of the same logic: send a GET request to a certain endpoint;
    if a requests.HTTPError arises, return None; else, transform the
    JSON response into an instance of a subclass of JSONWizard."""

    def function(s, e, i, u, h, p, sc, cf):
        url: str = f"{TIDAL_API_URL}/{e}/{i}{u}"
        kwargs: dict = {"url": url}
        if p is not None:
            kwargs["params"] = p
        if h is not None:
            kwargs["headers"] = h

        @backoff.on_predicate(
            backoff.expo,
            predicate=lambda r: r.status_code == 429,
            jitter=backoff.random_jitter,
            max_time=10,
            logger=logger,
        )
        def _get(s: Session, request_kwargs: dict) -> Response:
            """Return a requests.Response object, from having passed request_kwargs
            to s.get(), optionally retrying if 429 error occurs."""
            with s.get(**request_kwargs) as r:
                return r

        data: Optional[sc] = None
        logger.info(f"Requesting from TIDAL API: {e}/{i}{u}")
        resp: Response = _get(s=s, request_kwargs=kwargs)

        try:
            resp.raise_for_status()
        except HTTPError as he:
            if resp.status_code == 404:
                logger.warning(
                    f"404 Client Error: not found for TIDAL API endpoint {e}/{i}{u}"
                )
            elif resp.status_code == 401:
                logger.warning(
                    f"401 Client Error: Unauthorized for TIDAL API endpoint {e}/{i}{u}"
                )
            else:
                logger.exception(he)
        else:
            if cf:
                data = sc.from_dict({"credits": resp.json()})
            else:
                data = sc.from_dict(resp.json())
        finally:
            return data

    return function(
        s=session,
        e=endpoint,
        i=identifier,
        u=url_end,
        h=headers,
        p=parameters,
        sc=subclass,
        cf=credits_flag,
    )


# Functions will curry all arguments except session and identifier, as those
# don't change during runtime, and session is only available once the __main__
# process (on the happy path) creates a requests.Session object. Identifier
# varies with the media type, etc.

request_albums: Callable[
    [Session, int], Optional[AlbumsEndpointResponseJSON]
] = partial(
    requester_maker,
    endpoint="albums",
    headers={"Accept": "application/json"},
    subclass=AlbumsEndpointResponseJSON,
)

request_album_items: Callable[
    [Session, int], Optional[AlbumsItemsResponseJSON]
] = partial(
    requester_maker,
    endpoint="albums",
    headers={"Accept": "application/json"},
    parameters={"limit": 100},
    url_end="/items",
    subclass=AlbumsItemsResponseJSON,
)

request_album_review: Callable[
    [Session, int], Optional[AlbumsItemsResponseJSON]
] = partial(
    requester_maker,
    endpoint="albums",
    headers={"Accept": "application/json"},
    url_end="/review",
    subclass=AlbumsReviewResponseJSON,
)

request_artist_bio: Callable[
    [Session, int], Optional[ArtistsBioResponseJSON]
] = partial(
    requester_maker,
    endpoint="artists",
    headers={"Accept": "application/json"},
    url_end="/bio",
    subclass=ArtistsBioResponseJSON,
)

request_artists: Callable[
    [Session, int], Optional[ArtistsEndpointResponseJSON]
] = partial(
    requester_maker,
    endpoint="artists",
    headers={"Accept": "application/json"},
    subclass=ArtistsEndpointResponseJSON,
)

request_artists_albums: Callable[
    [Session, int], Optional[ArtistsAlbumsResponseJSON]
] = partial(
    requester_maker,
    endpoint="artists",
    headers={"Accept": "application/json"},
    url_end="/albums",
    subclass=ArtistsAlbumsResponseJSON,
)

request_artists_audio_works: Callable[
    [Session, int], Optional[ArtistsAlbumsResponseJSON]
] = partial(
    requester_maker,
    endpoint="artists",
    headers={"Accept": "application/json"},
    parameters={"filter": "EPSANDSINGLES"},
    url_end="/albums",
    subclass=ArtistsAlbumsResponseJSON,
)

request_artists_videos: Callable[
    [Session, int], Optional[ArtistsAlbumsResponseJSON]
] = partial(
    requester_maker,
    endpoint="artists",
    headers={"Accept": "application/json"},
    url_end="/videos",
    subclass=ArtistsVideosResponseJSON,
)

request_tracks: Callable[
    [Session, int], Optional[TracksEndpointResponseJSON]
] = partial(
    requester_maker,
    endpoint="tracks",
    headers={"Accept": "application/json"},
    subclass=TracksEndpointResponseJSON,
)

# This one's special, because its JSON response isn't proper JSON:
# it's just an array of JSON objects, so we have to pass a flag to mark
# that the logic common to the rest of the functions is slightly different here.
request_credits: Callable[
    [Session, int], Optional[TracksCreditsResponseJSON]
] = partial(
    requester_maker,
    endpoint="tracks",
    headers={"Accept": "application/json"},
    parameters={"includeContributors": True},
    url_end="/credits",
    subclass=TracksCreditsResponseJSON,
    credits_flag=True,
)

request_lyrics: Callable[[Session, int], Optional[TracksLyricsResponseJSON]] = partial(
    requester_maker,
    endpoint="tracks",
    headers={"Accept": "application/json"},
    url_end="/lyrics",
    subclass=TracksLyricsResponseJSON,
)

# One more layer of currying here, as the parameters argument
# is dependent on a runtime variable.
request_stream: Callable[
    [Session, int, str], Optional[TracksEndpointStreamResponseJSON]
] = lambda session, track_id, audio_quality: partial(
    requester_maker,
    session=session,
    identifier=track_id,
    endpoint="tracks",
    headers={"Accept": "application/json"},
    parameters={
        "audioquality": audio_quality,
        "playbackmode": "STREAM",
        "assetpresentation": "FULL",
    },
    url_end="/playbackinfopostpaywall",
    subclass=TracksEndpointStreamResponseJSON,
)()

request_videos: Callable[
    [Session, int], Optional[VideosEndpointResponseJSON]
] = partial(
    requester_maker,
    endpoint="videos",
    headers={"Accept": "application/json"},
    subclass=VideosEndpointResponseJSON,
)

request_video_contributors: Callable[
    [Session, int], Optional[VideosContributorsResponseJSON]
] = partial(
    requester_maker,
    endpoint="videos",
    headers={"Accept": "application/json"},
    parameters={"limit": 100},
    url_end="/contributors",
    subclass=VideosContributorsResponseJSON,
)

# One more layer of currying here, as the parameters argument
# is dependent on a runtime variable.
request_video_stream: Callable[
    [Session, int, str], Optional[VideosEndpointStreamResponseJSON]
] = lambda session, video_id, video_quality: partial(
    requester_maker,
    session=session,
    identifier=video_id,
    endpoint="videos",
    headers={"Accept": "application/json"},
    parameters={
        "videoquality": video_quality,
        "playbackmode": "STREAM",
        "assetpresentation": "FULL",
    },
    url_end="/playbackinfopostpaywall",
    subclass=VideosEndpointStreamResponseJSON,
)()

request_playlists: Callable[
    [Session, int], Optional[PlaylistsEndpointResponseJSON]
] = partial(
    requester_maker,
    endpoint="playlists",
    headers={"Accept": "application/json"},
    subclass=PlaylistsEndpointResponseJSON,
)


def get_album_id(session: Session, track_id: int) -> Optional[int]:
    """Given the Tidal ID to a track, query the Tidal API in order to retrieve
    the Tidal ID of the album to which the track belongs"""
    terj: Optional[TracksEndpointResponseJSON] = request_tracks(
        session=session, identifier=track_id
    )
    album_id: Optional[int] = None

    try:
        album_id = terj.id
    except AttributeError:
        pass
    finally:
        return album_id


def contiguous_ranges(value: int, range_size: int) -> Iterator[Tuple[int]]:
    """This function is a generator: it yields two-tuples of int, with the
    tuples representing the (inclusive) boundaries of ranges of size
    range_size. The final tuple will represent a range <= range_size if
    range_size does not evenly divide value. E.g.
    ```>>> list(ranges(16, 3))
    [(0, 2), (3, 5), (6, 8), (9, 11), (12, 14), (15, 16)]
    ```
    N.b. the first tuple will always have first element 0 and the final tuple
    will always have second element `value`."""
    i: int = 0
    rs: int = range_size - 1
    while i + rs < value:
        t: Tuple[int] = (i, i + rs)
        i = t[-1] + 1
        yield t
    else:
        yield (i, value)


def http_request_range_headers(
    content_length: int, range_size: int, return_tuple: bool = True
) -> Iterable[str]:
    """This function creates HTTP request Range headers. Its iterable
    returned is of tuples; each tuple describes the (inclusive) boundaries
    of a bytes range with size range_size. If return_tuple is False, it returns
    a generator of tuples. E.g.
    ```>>> http_request_range_headers(16, 3)
    ('bytes=0-2',
     'bytes=3-5',
     'bytes=6-8',
     'bytes=9-11',
     'bytes=12-14',
     'bytes=15-16')
    ```
    """
    ranges: Iterator[Tuple[int]] = contiguous_ranges(content_length, range_size)
    iterable: Iterable = (f"bytes={t[0]}-{t[1]}" for t in ranges)
    if return_tuple:
        return tuple(iterable)
    else:
        return iterable


def fetch_content_length(session: Session, url: str) -> dict:
    """Attempt to get the amount of bytes pointed to by `url`. If
    the HEAD request from the requests.Session object, `session`,
    encounters an HTTP request; or if the server does not support
    HTTP range requests; or if the server does not response with a
    Content-Length header, return 0"""
    session_params: dict = session.params
    # Unset params to avoid 403 response
    _params: dict = {k: None for k in session_params}
    with session.head(url=url, params=_params) as resp:
        if not resp.ok:
            cl: str = "0"
        else:
            cl: str = resp.headers.get("Content-Length", "0")
    return int(cl)



from dataclasses import dataclass, field
import json
import logging
from pathlib import Path
import shutil
import sys
from types import SimpleNamespace
from typing import Dict, List, Optional, Tuple, Union

from requests import HTTPError, Session

from .album import Album
from .media import AudioFormat
from .requesting import (
    request_artists,
    request_artists_albums,
    request_artists_audio_works,
    request_artists_videos,
)
from .track import Track
from .utils import download_cover_image, TIDAL_API_URL
from .video import Video

logger = logging.getLogger("__name__")


@dataclass
class Artist:
    artist_id: int

    def set_metadata(self, session: Session):
        """This function requests from TIDAL API endpoint /artists and
        stores the results in self.metadata"""
        self.metadata: Optional[ArtistsEndpointResponseJSON] = request_artists(
            session=session, identifier=self.artist_id
        )

    def save_artist_image(self, session: Session):
        artist_image: Path = self.artist_dir / "cover.jpg"
        if not artist_image.exists():
            download_cover_image(
                session, self.metadata.picture, self.artist_dir, dimension=750
            )

    def set_albums(self, session: Session):
        """This function requests from TIDAL API endpoint /artists/albums and
        stores the results in self.albums"""
        self.albums: Optional[ArtistsAlbumsResponseJSON] = request_artists_albums(
            session=session, identifier=self.artist_id
        )

    def set_audio_works(self, session: Session):
        """This function requests from TIDAL API endpoint
        /artists/albums?filter=EPSANDSINGLES and stores the results in self.albums"""
        self.albums: Optional[ArtistsAlbumsResponseJSON] = request_artists_audio_works(
            session=session, identifier=self.artist_id
        )

    def set_videos(self, session: Session):
        """This function requests from TIDAL API endpoint /artists/videos and
        stores the results in self.albums"""
        self.videos: Optional[ArtistsVideosResponseJSON] = request_artists_videos(
            session=session, identifier=self.artist_id
        )

    def set_dir(self, out_dir: Path):
        """This method sets self.artist_dir and creates the directory on the file system
        if it does not exist"""
        self.name: str = self.metadata.name.replace("..", "")
        self.artist_dir = out_dir / self.name
        self.artist_dir.mkdir(parents=True, exist_ok=True)

    def get_albums(
        self,
        session: Session,
        audio_format: AudioFormat,
        out_dir: Path,
        include_eps_singles: bool = False,
    ) -> List[Optional[str]]:
        if include_eps_singles:
            self.set_audio_works(session)
            logger.info(
                f"Starting attempt to get {self.albums.total_number_of_items} "
                "albums, EPs, and singles for artist with ID "
                f"{self.metadata.id},  '{self.name}'"
            )
        else:
            self.set_albums(session)
            logger.info(
                f"Starting attempt to get {self.albums.total_number_of_items} albums "
                f"for artist with ID {self.metadata.id}, '{self.name}'"
            )

        for i, a in enumerate(self.albums.items):
            album: Album = Album(album_id=a.id)
            album.get(
                session=session,
                audio_format=audio_format,
                out_dir=out_dir,
                metadata=a,
            )

    def get_videos(
        self,
        session: Session,
        out_dir: Path,
    ) -> List[Optional[str]]:
        """This method sets self.videos by calling self.set_videos()
        then, for each video, instantiates a Video object and executes
        video.get()"""
        self.set_videos(session)
        logger.info(
            f"Starting attempt to get {self.videos.total_number_of_items} videos "
            f"for artist with ID {self.metadata.id}, '{self.name}'"
        )
        for i, v in enumerate(self.videos.items):
            video: Video = Video(video_id=v.id)
            video.get(
                session=session,
                out_dir=out_dir,
                metadata=v,
            )

    def get(
        self,
        session: Session,
        audio_format: AudioFormat,
        out_dir: Path,
        include_eps_singles: bool,
    ):
        self.set_metadata(session)
        self.set_dir(out_dir)
        self.save_artist_image(session)
        self.get_videos(session, out_dir)
        self.get_albums(session, audio_format, out_dir, include_eps_singles)
